/**
 * Keeps track of all accelerators in use by the app and their purposes, whereas `globalShortcut.isRegistered`
 * only tracks shortcuts registered using `globalShortcut.register` and not via the application menu.
 *
 * This module only does bookkeeping--it does not actually create listeners for these accelerators.
 */
var ShortcutRegistry = {
  _registry: {},

  /**
   * Registers an accelerator for the specified purpose if one is not already registered.
   *
   * This method does not create a listener for this accelerator--you must still call
   * `globalShortcut.register` or `Menu.setApplicationMenu` separately.
   *
   * @param {String} accelerator - An accelerator.
   * @param {String} label - A label that describes the accelerator's purpose, like the name of the
   *   menu item with which it is associated.
   *
   * @return {Boolean} `true` if the registration was created, `false` if this accelerator was
   *   already registered for another purpose.
   */
  register: function(accelerator, label) {
    var registration = this.getRegistration(accelerator);
    if (registration) return false;

    this._registry[accelerator] = {
      accelerator: accelerator,
      label: label
    };

    return true;
  },

  unregister: function(accelerator) {
    delete this._registry[accelerator];
  },

  /**
   * Gets any existing registration for the specified accelerator.
   *
   * @param {String} accelerator - An accelerator.
   *
   * @return {Object}
   *   @property {String} accelerator - The accelerator.
   *   @property {String} label - The label with which the accelerator was registered.
   */
  getRegistration: function(accelerator) {
    return this._registry[accelerator];
  }
};

module.exports = ShortcutRegistry;
