var _ = require('underscore');
var app = require('electron').app;
var EventEmitter = require('events').EventEmitter;
var ipc = require('ipc-main');
var fs = require('fs');
var path = require('path');
var util = require('util');

/**
 * We must cache relevant user preferences app-side so that we can apply them before Meteor starts up.
 */
var DEFAULT_PREFERENCES_PATH = path.join(__dirname, 'defaultUserPreferences.json');
var USER_PREFERENCES_PATH = path.join(app.getPath('userData'), 'userPreferences.json');

var UserPreferencesModel = function() {
  EventEmitter.call(this);

  try {
    this._preferences = require(USER_PREFERENCES_PATH);
  } catch(e) {
    this._preferences = require(DEFAULT_PREFERENCES_PATH);
    this._syncPreferencesToDisk();
  }
};

util.inherits(UserPreferencesModel, EventEmitter);

_.extend(UserPreferencesModel.prototype, {
  _syncPending: false,
  _syncInProgress: false,

  set: function(keyOrPrefs, value /* optional if keyOrPrefs is an object */) {
    var prefs = keyOrPrefs;
    if (!_.isObject(prefs)) {
      prefs = {};
      prefs[keyOrPrefs] = value;
    }

    var changes = [];
    _.each(prefs, function(val, key) {
      if (!_.isEqual(this._preferences[key], val)) {
        changes.push(key);
        this._preferences[key] = val;
      }
    }, this);

    if (!_.isEmpty(changes)) {
      _.each(changes, function(key) {
        this.emit('change:' + key, this._preferences[key]);
      }, this);
      this.emit('change');

      this._syncPreferencesToDisk();
    }
  },

  get: function(key) {
    return this._preferences[key];
  },

  mergeRemoteUpdates: function(window) {
    ipc.on('preferencesUpdated', function(event, prefs) {
      if (event.sender === window.webContents) {
        this.set(prefs);
      }
    }.bind(this));
  },

  _syncPreferencesToDisk: function() {
    if (this._syncInProgress) {
      this._syncPending = true;
      return;
    }

    this._syncPending = false;
    this._syncInProgress = true;
    fs.writeFile(USER_PREFERENCES_PATH, JSON.stringify(this._preferences), function(err) {
      this._syncInProgress = false;

      if (err) {
        // TODO(jeff): Add Sentry logging.
      }
      if (this._syncPending) this._syncPreferencesToDisk();
    }.bind(this));
  }
});

module.exports = new UserPreferencesModel();
