/* global ElectronImplementation:true */

/**
 * Since we've disabled Node integration in the browser window, we must selectively expose
 * main-process/Node modules via this script as well as patch the browser environment in certain ways.
 *
 * @WARNING This file must take care not to leak the imported modules to the browser window!
 * In particular, do not save the following variables as properties of `ElectronImplementation`.
 * See https://github.com/atom/electron/issues/1753#issuecomment-104719851.
 *
 * @WARNING Do not attempt to use main process modules, in particular our *own* modules, from the
 * browser window--they will not work! Even if they don't rely on Node functionality, they will not
 * share a JavaScript context with the main process--methods must use a proxy as is done with
 * `shortcutRegistry` below.
 *
 * @WARNING When passing callbacks to remote objects, be careful about leaks as described in
 * https://github.com/atom/electron/blob/master/docs/api/remote.md#passing-callbacks-to-the-main-process.
 */
var _ = require('underscore');
var ipc = require('electron').ipcRenderer;
var Environment = require('../../utils/Environment');
var path = require('path');
var remote = require('electron').remote;
var shell = require('electron').shell;
var webFrame = require('electron').webFrame;

// Main process modules, either because they are shared between the main and renderer processes or
// because they themselves use main process modules; we must load them via the remote proxy for
// reasons described at top. We must provide absolute paths to `remote.require`.
var buildEditorContextMenu = remote.require('electron-editor-context-menu');
var shortcutRegistry = remote.require(path.resolve(__dirname, '../../utils/shortcutRegistry'));

// TODO(jeff): Figure out how to build this on Windows.
var SpellCheckProvider;
if (Environment.isMac()) {
  // I'm not sure if this should technically be required by the remote too (even though it doesn't fit
  // the previous definition of a main process module), but I don't want to because of fear of leaking
  // event listeners as described at top.
  SpellCheckProvider = require('electron-spell-check-provider');
}

/**
 * Defines methods with which to extend the `Electron` module defined in `client.js`.
 * This must be a global in order to escape the preload script and be available to `client.js`.
 */
ElectronImplementation = {
  /**
   * Open the given external protocol URL in the desktop's default manner. (For example, http(s):
   * URLs in the user's default browser.)
   *
   * @param {String} url - The URL to open.
   */
  openExternal: shell.openExternal,

  /**
   * Determines if the browser window is currently in fullscreen mode.
   *
   * "Fullscreen" here refers to the state triggered by toggling the native controls, not that
   * toggled by the HTML API.
   *
   * To detect when the browser window changes fullscreen state, observe the 'enter-full-screen'
   * and 'leave-full-screen' events using `onWindowEvent`.
   *
   * @return {Boolean} `true` if the browser window is in fullscreen mode, `false` otherwise.
   */
  isFullScreen: function() {
    return remote.getCurrentWindow().isFullScreen();
  },

  /**
   * Shows a window previously hidden e.g. by opening it using 'show=0'.
   *
   * @param {String} id - The ID of a window. The ID of a window opened using `var win = window.open`
   *   is available as `win.guestId`.
   */
  showWindow: function(id) {
    remote.BrowserWindow.fromId(id).show();
  },

  /**
   * Sends a message to Electron.
   *
   * @param {String} method - The name of the message.
   * @param {JSON=} payload - (Optional) A JSON-serializable payload.
   */
  send: function(method, payload) {
    ipc.send(method, payload);
  },

  /**
   * Invokes _callback_ when the specified `BrowserWindow` event is fired.
   *
   * This differs from `onEvent` in that it directs Electron to start emitting the relevant window
   * event.
   *
   * See https://github.com/atom/electron/blob/master/docs/api/browser-window.md#events for a list
   * of events.
   *
   * The implementation of this API, in particular the use of the `ipc` vs. `remote` modules, is
   * designed to avoid memory leaks as described by
   * https://github.com/atom/electron/blob/master/docs/api/remote.md#passing-callbacks-to-the-main-process.
   *
   * @param {String} event - The name of a `BrowserWindow` event.
   * @param {Function} callback - A function to invoke when `event` is triggered. Takes no arguments
   *   and returns no value.
   */
  onWindowEvent: function(event, callback) {
    this.onEvent(event, callback);
    this.send('observe-window-event', event);
  },

  /**
   * Invokes _callback_ when the specified IPC event is fired.
   *
   * @param {String} event - The name of an event.
   * @param {Function} callback - A function to invoke when `event` is triggered. Takes no arguments
   *   and returns no value.
   */
  onEvent: function(event, callback) {
    var listeners = this._eventListeners[event];
    if (!listeners) {
      listeners = this._eventListeners[event] = [];
      ipc.on(event, function(/* event, ... */) {
        // Ignore the event argument.
        var args = _.rest(arguments);
        _.invoke(listeners, 'apply', null, args);
      });
    }
    listeners.push(callback);
  },

  /**
   * Gets any existing registration for the specified accelerator.
   *
   * @param {String} accelerator - An accelerator.
   *
   * @return {Object}
   *   @property {String} accelerator - The accelerator.
   *   @property {String} label - The label with which the accelerator was registered.
   */
  getShortcutRegistration: function(accelerator) {
    return shortcutRegistry.getRegistration(accelerator);
  },

  _eventListeners: {}
};


// Make sure that new windows are shown unless specified otherwise, to work around
// https://github.com/atom/electron/issues/3751.
// This is necessary for us to preload compose moles hidden, as well as to hide the main window until
// it's loaded.
(function() {
  /* jshint browser: true */
  var oldWindowOpen = window.open;

  window.open = function(url, windowName, features) {
    if (!windowName) windowName = '';

    if (!features || (features.indexOf('show=') < 0)) {
      features = (features ? (features + ',') : '') + 'show=1';
    }

    return oldWindowOpen.call(window, url, windowName, features);
  };
})();


// Enable spell-checking and the right-click context menu in text editors.
// Electron (`webFrame.setSpellCheckProvider`) only underlines misspelled words;
// we must manage the menu ourselves.
(function() {
  // Track information about the selection to show in the context menu.
  var selection;
  function resetSelection() {
    selection = {
      isMisspelled: false,
      spellingSuggestions: []
    };
  }
  resetSelection();

  // Reset the selection when clicking around, before the spell-checker runs and the context menu shows.
  window.addEventListener('mousedown', resetSelection);

  // TODO(jeff): Figure out how to build the spell-checker on Windows.
  if (Environment.isMac()) {
    // The spell-checker runs when the user clicks on text and before the 'contextmenu' event fires.
    // Thus, we may retrieve spell-checking suggestions to put in the menu just before it shows.
    webFrame.setSpellCheckProvider(
      'en-US',
      // Not sure what this parameter (`autoCorrectWord`) does: https://github.com/atom/electron/issues/4371
      // The documentation for `webFrame.setSpellCheckProvider` passes `true` so we do too.
      true,
      new SpellCheckProvider('en-US').on('misspelling', function(suggestions) {
        // Prime the context menu with spelling suggestions _if_ the user has selected text. Electron
        // may sometimes re-run the spell-check provider for an outdated selection e.g. if the user
        // right-clicks some misspelled text and then an image.
        if (window.getSelection().toString()) {
          selection.isMisspelled = true;
          selection.spellingSuggestions = _.first(suggestions, 3);
        }
      }));
  }

  window.addEventListener('contextmenu', function(e) {
    // Only show the context menu in text editors.
    if (!e.target.closest('textarea, input, [contenteditable="true"]')) return;

    var menu = buildEditorContextMenu(selection);

    // The 'contextmenu' event is emitted after 'selectionchange' has fired but possibly before the
    // visible selection has changed. Try to wait to show the menu until after that, otherwise the
    // visible selection will update after the menu dismisses and look weird.
    setTimeout(function() {
      menu.popup(remote.getCurrentWindow());
    }, 30);
  });
})();
