var _ = require('underscore');
var app = require('electron').app;
var BrowserWindow = require('electron').BrowserWindow;
var Environment = require('../../utils/Environment');
var Menu = require('electron').Menu;

/**
 * Guards against the main window being shown when the menu bar icon messages it to show a compose
 * window, and when the compose windows are closed.
 */
var preventMainWindowFromBeingShown = function(mainWindow) {
  if (Environment.isWindows()) {
    // Unlike on Mac, the main window does not become visible when the menu bar icon messages it,
    // nor when the compose windows are closed.
    //
    // TODO(jeff): Make the previously-focused app become focused again when the compose windows are closed:
    // https://app.asana.com/0/62901882894067/77337079005097
    return {
      preClickHandler: _.noop,
      postClickHandler: _.noop
    };
  }


  // Gets all windows ignoring the preloaded compose moles.
  function getAllUserWindows() {
    return _.filter(BrowserWindow.getAllWindows(), function(window) {
      return window.isVisible() || (window === mainWindow);
    });
  }

  // This gives focus to the next application like if the user hit Cmd-H, vs. `BrowserWindow.hide`
  // which just hides the main window but keeps the app active: https://github.com/atom/electron/issues/2640
  function hideApp() {
    // If the main window is not visible when 'hide:' is sent, the app will not relinquish event
    // listeners: https://app.asana.com/0/19141607276671/77224967483359
    // Showing the main window may flash it--but only rarely, for some reason, thankfully.
    //
    // (To prevent the flash: showing it offscreen doesn't work because `BrowserWindow#show` moves
    // it back onscreen; showing it at a small size is unfavorable because when we *actually* show
    // it next, it'll "pop" as the browser re-renders it at the new size (re-rendering only occurs
    // when visible).)
    if (!mainWindow.isVisible()) mainWindow.show();
    Menu.sendActionToFirstResponder('hide:');
  }

  var mainWindowWasVisible = false;
  var mainWindowWasFocusedBeforeOrSinceOpeningAppIcon = false;

  var preClickHandler = function() {
    mainWindowWasVisible = mainWindow.isVisible();

    mainWindowWasFocusedBeforeOrSinceOpeningAppIcon = mainWindow.isFocused() ||
      // Accounts for the icon being clicked again before closing the first window.
      (mainWindowWasFocusedBeforeOrSinceOpeningAppIcon && (_.size(getAllUserWindows()) > 1));
  };

  var postClickHandler = function() {
    // Sending a message will have caused the main window to become visible if it wasn't. Stop that.
    if (!mainWindowWasVisible) {
      mainWindow.hide();
    }
  };

  mainWindow.on('focus', function() {
    mainWindowWasFocusedBeforeOrSinceOpeningAppIcon = true;
  });

  app.on('browser-window-created', function(e, win) {
    win.on('closed', function() {
      // When the user closes the last secondary window, the app will try to return focus to the main window.
      // But if the app didn't have focus before the window was opened, we should instead hide the
      // app altogether.
      if ((_.size(getAllUserWindows()) === 1) && !mainWindowWasFocusedBeforeOrSinceOpeningAppIcon) {
        hideApp();
      }
    });
  });

  return {
    preClickHandler: preClickHandler,
    postClickHandler: postClickHandler
  };
};

module.exports = preventMainWindowFromBeingShown;
